# -*- coding: utf-8 -*-
#
# Copyright (c) 2016 Muges
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation files
# (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge,
# publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""
Formatting functions
"""


def format_duration(seconds):
    """Format a duration as ``[hours:]minutes:seconds``.

    Parameters
    ----------
    seconds : int
        Duration in seconds.

    Returns
    -------
    str
    """
    minutes, seconds = divmod(seconds, 60)
    hours, minutes = divmod(minutes, 60)

    if hours > 0:
        return "{}:{:02d}:{:02d}".format(hours, minutes, seconds)
    else:
        return "{}:{:02d}".format(minutes, seconds)


def format_fulltext_duration(seconds):
    """Format a duration as ``h hours, m minutes``.

    Parameters
    ----------
    seconds : int
        Duration in seconds.

    Returns
    -------
    str
    """
    minutes, seconds = divmod(seconds, 60)
    hours, minutes = divmod(minutes, 60)

    if hours > 2:
        if minutes >= 30:
            hours += 1
        return "{} hours".format(hours)
    elif hours > 0:
        return "{} hours, {} minutes".format(hours, minutes)
    if minutes > 2:
        if seconds >= 30:
            minutes += 1
        return "{} minutes".format(minutes)
    elif minutes > 0:
        return "{} minutes, {} seconds".format(minutes, seconds)
    else:
        return "{} seconds".format(seconds)


def format_size(size):
    """Convert a size in bytes into a human readable string.

    Example
    -------
    ::

        >>>format_size(2.25*1024)
        '2.25 kB'

    Parameters
    ----------
    size : int
        Size in bytes.

    Returns
    -------
    str
    """
    for unit in ['bytes', 'kB', 'MB']:
        if size < 1024:
            return "{:.2f} {}".format(size, unit)
        size /= 1024

    return "{:.2f} {}".format(size, 'GB')



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (c) 2016 Muges
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation files
# (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge,
# publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# pylint: disable=invalid-name

import os
import sys
from setuptools import setup, find_packages

from erika.__version__ import __appname__, __version__

# 'setup.py publish' shortcut.
if sys.argv[-1] == 'publish':
    os.system('python setup.py sdist bdist_wheel')
    os.system('twine upload dist/*')
    sys.exit()

with open('README.rst') as f:
    long_description = f.read()


def data_directory(dst, src):
    """Generate a list of ``(directory, files)`` pairs to be used in the
    ``data_files`` parameter of the ``setup`` function in order to move the
    content of the ``src`` directory in the ``dst`` directory."""
    data_files = []

    for root, _, filenames in os.walk(src):
        if not filenames:
            continue

        filenames = [
            os.path.join(root, filename) for filename in filenames
        ]
        dirname = os.path.relpath(root, src)
        data_files.append((os.path.join(dst, dirname), filenames))

    return data_files


setup(
    name=__appname__.lower(),
    version=__version__,
    description='A GTK podcast manager',
    long_description=long_description,
    license='MIT',
    url='https://github.com/Muges/erika',
    author='Muges',
    author_email='git@muges.fr',

    packages=find_packages(),
    package_data={
        'erika.frontend': [
            'data/*.glade',
            'data/*.html'
        ],
    },
    data_files=[
        ('share/erika', ['README.rst', 'CHANGELOG.rst']),
        ('share/applications', ['data/erika.desktop']),
    ] + data_directory('share/icons', 'data/icons'),

    install_requires=[
        'feedparser',
        'lxml',
        'mutagen',
        'mygpoclient',
        'peewee',
        'pillow',
        'requests',
    ],
    tests_require=[
        'pytest'
    ],

    entry_points={
        'console_scripts': [
            'erika = erika.frontend:run',
        ],
    },

    classifiers=[
        'Environment :: X11 Applications :: GTK',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3 :: Only',
        'Topic :: Multimedia :: Sound/Audio :: Players'
    ]
)



# -*- coding: utf-8 -*-
#
# Copyright (c) 2016 Muges
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation files
# (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge,
# publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""
A model used to store podcasts in the library.
"""

import logging
import os.path
import requests
from peewee import BooleanField, TextField, DoesNotExist, IntegrityError, fn

from erika import parsers
from erika.library.fields import Image, ImageField
from erika.util import sanitize_filename
from .database import BaseModel, database, slugify
from .config import Config
from .episode import Episode
from .episode import Podcast as PodcastProxy
from .podcast_action import PodcastAction


class Podcast(BaseModel):
    """A model used to store podcasts in the library.

    Attributes
    ----------
    parser : str
        The name of the parser (see :mod:`erika.parsers`).
    url : str
        The url of the podcast.
    title : str, optional
        The title of the podcast.
    author : str, optional
        The author of the podcast.
    image_url : str, optional
        A link to the image of the podcast
    image : :class:`~erika.library.fields.Image`, optional
        The image of the podcast.
    language : str, optional
        The language of the podcast.
    subtitle : str, optional
        The subtitle of the podcast.
    summary : str, optional
        A description of the podcast.
    link : str, optional
        Website link.
    update_failed : bool
        True if the last update of the podcast failed.
    """
    parser = TextField()
    url = TextField()

    title = TextField(null=True)
    author = TextField(null=True)
    image_url = TextField(null=True)
    image = ImageField(null=True)
    language = TextField(null=True)
    subtitle = TextField(null=True)
    summary = TextField(null=True)
    link = TextField(null=True)

    update_failed = BooleanField(default=False)

    class Meta:  # pylint: disable=too-few-public-methods, missing-docstring
        indexes = (
            (('parser', 'url'), True),
        )

    @property
    def display_subtitle(self):
        """str: The subtitle of the podcast, or its summary if it has no
        subtitle."""
        return self.subtitle or self.summary or ""

    @property
    def display_summary(self):
        """str: The summary of the podcast, or its subtitle if it has no
        summary."""
        return self.summary or self.subtitle or ""

    @property
    def display_title(self):
        """str: The title of the podcast, or its url if it has no title."""
        return self.title or self.url or ""

    @property
    def episodes_count(self):
        """int: the number of episodes of the podcast."""
        return self.episodes.count()

    @property
    def played_count(self):
        """int: the number of episodes that have already been played."""
        # pylint: disable=singleton-comparison
        return self.episodes.where(Episode.played == True).count()

    @property
    def new_count(self):
        """int: the number of episodes that have been added since the previous
        launch."""
        # pylint: disable=singleton-comparison
        return self.episodes.where(Episode.new == True).count()

    @property
    def unplayed_count(self):
        """int: the number of episodes that have not been played."""
        return self.episodes_count - self.played_count

    @classmethod
    def new(cls, parser, url):
        """Create a new podcast, and add it to the database

        Parameters
        ----------
        parser_name : str
            The name of a parser (see :mod:`erika.parsers`).
        url : str
            The url of the podcast.
        """
        logger = logging.getLogger(".".join((__name__, cls.__name__)))
        logger.info("Adding the %s podcast %s.", parser, url)
        try:
            podcast = Podcast.get(parser=parser, url=url)
        except DoesNotExist:
            podcast = Podcast(parser=parser, url=url)
            podcast.save()

            if podcast.parser == 'rss':
                PodcastAction.new(podcast_url=podcast.url, action="add")
        else:
            logger.warning("The %s podcast %s is already in the database.",
                           parser, url)

        return podcast

    def delete_instance(self, *args, **kwargs):
        # pylint: disable=arguments-differ
        PodcastAction.new(podcast_url=self.url, action="remove")
        super().delete_instance(*args, **kwargs)

    def update_podcast(self):
        """Update the podcast."""
        logger = logging.getLogger(
            ".".join((__name__, self.__class__.__name__)))
        logger.info("Updating the podcast %s.", self.display_title)

        try:
            # Parse the podcast
            episodes = parsers.parse(self)

            # Download the podcast's image if its url changed
            dirty_fields = [field.name for field in self.dirty_fields]
            if 'image_url' in dirty_fields:
                self.download_image()

            # Get the track number that should be used for the next new episode
            next_track_number = self.get_next_track_number()

            # Save the podcast and the episodes in the database
            with database.transaction():
                self.update_failed = False
                self.save()

                for episode in episodes:
                    episode.track_number = next_track_number
                    try:
                        episode.save()
                    except IntegrityError:
                        # The episode is already in the database, don't
                        # increment the track number.
                        continue
                    next_track_number += 1
        except Exception:  # pylint: disable=broad-except
            logger.exception("Unable to update the podcast.")

            self.update_failed = True
            self.save()

    def download_image(self):
        """Download the podcast's image"""
        logger = logging.getLogger(
            ".".join((__name__, self.__class__.__name__)))
        logger.debug("Downloading image for the podcast %s.", self.title)
        try:
            response = requests.get(self.image_url)
            response.raise_for_status()
        except requests.exceptions.RequestException:
            logger.exception("Unable to download the image.")
            return

        self.image = Image(response.content)

    def get_next_track_number(self):
        """Get the track number of the next episode."""
        previous = self.episodes.select(fn.Max(Episode.track_number)).get()
        previous_number = previous.track_number or 0
        return previous_number + 1

    def get_directory(self):
        """Return the directory containing a podcast's audio files.

        Returns
        -------
        str
            The path of the directory.
        """
        template = Config.get_value("library.podcast_directory_template")
        return os.path.join(
            Config.get_value("library.root"),
            sanitize_filename(template.format(podcast=self))
        )

    def get_episode_from_tags_strict(self, audiotags):
        """Get the podcast's episode matching a file strictly.

        Returns the episode having the same guid, or the same title
        and publication date if there is one (meaning it was probably
        tagged by erika).

        Parameters
        ----------
        audiotags : dict
            Dictionnary containing the audio file's metadata.

        Returns
        -------
        :class:`Episode`, optional
            The episode, or None if there was no match.
        """
        logger = logging.getLogger(
            ".".join((__name__, self.__class__.__name__)))

        guid = audiotags['guid']
        pubdate = audiotags['pubdate']
        title = audiotags['title']

        try:
            episode = self.episodes.where(
                Episode.local_path.is_null(),
                (Episode.guid == guid) |
                ((Episode.pubdate == pubdate) & (Episode.title == title))
            ).first()
        except DoesNotExist:
            logger.debug("No strict match.")
            return None

        return episode

    def get_episode_from_tags_loose(self, audiotags, filename):
        """Get the podcast's episode matching a file loosely.

        Returns the episode having a title similar to the title or
        filename of the audio file. If there is more than one such
        match, None is returned as there is no way to find which
        episode is the right one.

        Parameters
        ----------
        audiotags : dict
            Dictionnary containing the audio file's metadata.
        filename : str
            The name of the audio file (without extensions or directory).

        Returns
        -------
        :class:`Episode`, optional
            The episode, or None if there was no match.
        """
        logger = logging.getLogger(
            ".".join((__name__, self.__class__.__name__)))

        title = audiotags['title']

        # Try to find a single episode with the same title
        episodes = list(self.episodes.where(
            Episode.local_path.is_null(),
            Episode.title == title).limit(2))
        if episodes:
            if len(episodes) > 1:
                logger.warning("Too many episodes with title %s.", title)
                return None

            return episodes[0]

        # Try to find a single episode with a title or filename similar to the
        # title
        filename = slugify(filename)
        title = slugify(title or "")

        episodes = list(self.episodes.where(
            Episode.local_path.is_null(),
            fn.slugify(Episode.title) << [title, filename]).limit(2))
        if episodes:
            if len(episodes) > 1:
                logger.warning("Too many episodes with title similar to '%s' "
                               "or '%s'.", title, filename)
                return None

            return episodes[0]


PodcastProxy.initialize(Podcast)



# -*- coding: utf-8 -*-
#
# Copyright (c) 2016 Muges
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation files
# (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge,
# publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""
The :mod:`erika.tags` module provides functions to set and get tags for various
audio format.

For now, the only supported format is MP3.
"""

import logging
import mutagen

from . import mp3


def set_tags(filename, episode):
    """Set the tags of the audio file of an episode.

    Parameters
    ----------
    filename : str
        The audio file.
    episode : :class:`Episode`
        The corresponding episode.
    """
    logger = logging.getLogger(__name__)

    try:
        audio = mutagen.File(filename)
    except Exception:  # pylint: disable=broad-except
        logger.exception("Unable to set tags for '%s'.", filename)
        return None

    if isinstance(audio, mutagen.mp3.MP3):
        mp3.set_tags(audio, episode)
    else:
        logger.warning("Unable to set tags (unsupported file format) for "
                       "'%s'.", filename)


def get_tags(filename):
    """Get the tags of an audio file.

    Parameters
    ----------
    filename : str
        The audio file.
    """
    logger = logging.getLogger(__name__)

    try:
        audio = mutagen.File(filename)
    except Exception:  # pylint: disable=broad-except
        logger.exception("Unable to get tags for '%s'.", filename)
        return None

    if isinstance(audio, mutagen.mp3.MP3):
        return mp3.get_tags(audio)
    else:
        logger.warning("Unable to get tags (unsupported file format) for "
                       "'%s'.", filename)
        return None



# -*- coding: utf-8 -*-
#
# Copyright (c) 2016 Muges
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation files
# (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge,
# publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""
A button used to represent network connectivity options
"""

from gi.repository import Gtk
from gi.repository import Gio

from erika.frontend.util import cb


class NetworkButton(Gtk.Button):
    """A button used to represent network connectivity options"""
    def __init__(self):
        Gtk.Button.__init__(self)
        self.set_relief(Gtk.ReliefStyle.NONE)

        self.online_image = Gtk.Image.new_from_icon_name(
            "network-transmit-receive-symbolic", Gtk.IconSize.SMALL_TOOLBAR)
        self.offline_image = Gtk.Image.new_from_icon_name(
            "network-offline-symbolic", Gtk.IconSize.SMALL_TOOLBAR)
        self.error_image = Gtk.Image.new_from_icon_name(
            "network-error-symbolic", Gtk.IconSize.SMALL_TOOLBAR)

        self.online = True
        self.error = False
        self.set_network_state(self.online, self.error)

        application = Gio.Application.get_default()
        application.connect('network-state-changed',
                            cb(self.set_network_state))

        self.connect('clicked',
                     lambda _: application.set_network_state(not self.online))

    def set_network_state(self, online, error):
        """Change the state of the button"""
        self.online = online
        self.error = error
        if self.online:
            self.set_image(self.online_image)
            self.set_tooltip_text("Online")
        elif self.error:
            self.set_image(self.error_image)
            self.set_tooltip_text("Unable to connect the internet")
        else:
            self.set_image(self.offline_image)
            self.set_tooltip_text("Offline mode")



# -*- coding: utf-8 -*-
#
# Copyright (c) 2016 Muges
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation files
# (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge,
# publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""
A toggle button used to represent filtering options.
"""

from gi.repository import Gtk
from gi.repository import Gdk


class FilterButton(Gtk.ToggleButton):
    """A toggle button used to represent filtering options

    The button has three possible states:
     - None or inactive (i.e. show everything);
     - True (e.g. show new episodes);
     - False (e.g. show non new episodes).

    Parameters
    ----------
    icon_name : str
        The name of the icon of the button
    name : str
        The name of the filter, it is used in the tooltip of the button ("Show
        all", "Show only {name}" or "Hide {name}").
    key : Any
        This is not used by the FilterButton itself, but can be used to store a
        filter key, which will then be accessible through the key attribute.
    """
    def __init__(self, icon_name, name, key=None):
        Gtk.ToggleButton.__init__(self)
        self.connect("button-press-event", FilterButton._on_button_press_event)

        self.name = name
        self.key = key
        self.state = None

        self.inactive_image = Gtk.Image.new_from_icon_name(
            icon_name, Gtk.IconSize.SMALL_TOOLBAR)
        self.filter_true_image = Gtk.Image.new_from_icon_name(
            icon_name, Gtk.IconSize.SMALL_TOOLBAR)
        self.filter_false_image = Gtk.Image.new_from_icon_name(
            icon_name, Gtk.IconSize.SMALL_TOOLBAR)
        self.filter_false_image.set_opacity(0.5)

        self._set_state(None)

    def _set_state(self, state):
        """Set the state of the button

        Parameters
        ----------
        state : Union[bool, None]
        """
        self.state = state
        self.set_active(state is not None)

        if self.state is None:
            self.set_image(self.inactive_image)
            self.set_tooltip_text("Show only {}".format(self.name))
        elif self.state:
            self.set_image(self.filter_true_image)
            self.set_tooltip_text("Hide {}".format(self.name))
        else:
            self.set_image(self.filter_false_image)
            self.set_tooltip_text("Show all")

    def _on_button_press_event(self, event):
        """Called when the button is clicked"""
        # Propagate the event if it is not a single left button click
        if event.type != Gdk.EventType.BUTTON_PRESS or event.button != 1:
            return False

        # Rotate through the states (None -> True -> False -> None...)
        if self.state is None:
            self._set_state(True)
        elif self.state:
            self._set_state(False)
            self.emit("toggled")
        else:
            self._set_state(None)

        return True



# -*- coding: utf-8 -*-
#
# Copyright (c) 2016 Muges
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation files
# (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge,
# publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""
Main window of the application
"""

from gi.repository import Gtk

from erika.__version__ import __appname__
from erika.library.models import Episode
from .details import Details
from .downloads import DownloadsButton
from .episode_list import EpisodeList
from .player import Player
from .player_widgets import PlayerTitle, PlayerControls
from .podcast_list import PodcastList
from .util import cb, get_builder
from .widgets import StatusBox, NetworkButton, Paned


class MainWindow(Gtk.ApplicationWindow):
    """
    Main window of the application
    """
    def __init__(self, application):
        Gtk.ApplicationWindow.__init__(self, application=application,
                                       title=__appname__)
        self.set_default_size(800, 600)
        self.connect("destroy", self._on_close)

        # Header bar
        menu_button = Gtk.MenuButton()
        menu_button.set_image(Gtk.Image.new_from_icon_name(
            "open-menu-symbolic", Gtk.IconSize.BUTTON))

        builder = get_builder('data/menu.glade')
        menu_button.set_menu_model(builder.get_object("app-menu"))

        self.downloads_button = DownloadsButton()
        self.downloads_button.connect("episode-updated",
                                      self._on_episode_updated)

        # Player
        self.player = Player()
        player_title = PlayerTitle(self.player)
        player_controls = PlayerControls(self.player)

        self.player.connect("episode-updated", self._on_episode_updated)
        self.player.connect('progress-changed', cb(player_title.set_progress))
        self.player.connect('state-changed',
                            cb(player_controls.set_player_state, 2))
        self.player.connect('state-changed',
                            cb(player_title.set_player_state))

        # Error bar
        self.error_bar = Gtk.InfoBar()
        self.error_bar.set_no_show_all(True)
        self.error_bar.set_message_type(Gtk.MessageType.ERROR)
        self.error_bar.set_show_close_button(True)
        self.error_bar.connect('response', cb(self.error_bar.hide, 2))

        self.error_label = Gtk.Label()
        self.error_label.show()
        self.error_bar.get_content_area().add(self.error_label)

        # Views
        self.details = Details()

        self.podcast_list = PodcastList()
        self.episode_list = EpisodeList(self.player)

        self.podcast_list.connect('podcast-selected',
                                  cb(self.episode_list.select))
        self.podcast_list.connect('podcast-selected',
                                  cb(self.details.show_podcast))

        self.episode_list.connect('episodes-changed',
                                  self._on_episodes_changed)
        self.episode_list.connect('download',
                                  cb(self.downloads_button.download))
        self.episode_list.connect('episode-selected',
                                  cb(self.details.show_episode))
        self.episode_list.connect('podcast-selected',
                                  cb(self.details.show_podcast))

        self.player.connect('progress-changed',
                            cb(self.episode_list.set_player_progress))
        self.player.connect('state-changed',
                            cb(self.episode_list.set_player_state))

        # Header bar
        headerbar = Gtk.HeaderBar()
        headerbar.set_custom_title(player_title)
        headerbar.set_show_close_button(True)
        headerbar.pack_start(menu_button)
        headerbar.pack_start(player_controls)
        headerbar.pack_end(self.downloads_button)
        headerbar.props.title = __appname__
        self.set_titlebar(headerbar)

        # Statusbar
        self.counts = Gtk.Label()
        self.statusbox = StatusBox()

        self._build_layout()

        self.update_counts()
        self.podcast_list.update()

    def _build_layout(self):
        vbox = Gtk.VBox()
        self.add(vbox)

        vbox.pack_start(self.error_bar, False, False, 0)

        right_paned = Paned()
        right_paned.set_limit_width(700)
        right_paned.set_position(800)
        right_paned.add1(self.episode_list)
        right_paned.add2(self.details)

        paned = Gtk.Paned()
        paned.set_position(300)
        vbox.pack_start(paned, True, True, 0)

        paned.add1(self.podcast_list)
        paned.add2(right_paned)

        # Status bar
        status_bar = Gtk.HBox()
        status_bar.set_margin_top(5)
        status_bar.set_margin_bottom(5)
        status_bar.set_margin_start(5)
        status_bar.set_margin_end(5)
        vbox.pack_start(status_bar, False, False, 0)

        status_bar.pack_start(NetworkButton(), False, False, 0)
        status_bar.pack_start(self.counts, False, False, 0)
        status_bar.pack_end(self.statusbox, False, False, 0)

    def update_counts(self):
        """
        Update counts label
        """
        new, played, total = Episode.get_counts()
        unplayed = total - played
        self.counts.set_text("{} new episodes, {} unplayed".format(
            new,
            unplayed
        ))

    def _on_close(self, window):
        """
        Called when the window is closed.
        """
        self.player.stop()
        self.downloads_button.stop()

    def _on_episodes_changed(self, episodes_list):
        """
        Called when episodes are modified (e.g. marked as played) in the
        episodes list.

        Update the current podcast.
        """
        self.podcast_list.update_current()
        self.update_counts()

    def _on_episode_updated(self, widget, episode):
        """
        Called when an episode is updated by a widget
        """
        self.podcast_list.update_podcast(episode.podcast)
        self.episode_list.update_episode(episode)



# -*- coding: utf-8 -*-
#
# Copyright (c) 2016 Muges
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation files
# (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge,
# publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""
Button displaying the list downloads and their progress
"""

import logging

from gi.repository import GObject
from gi.repository import Gtk
from gi.repository import GLib

from erika.downloads import DownloadsPool, download_with_average_speed
from erika.util import format_fulltext_duration, format_size
from .widgets import Label, ScrolledWindow, IndexedListBox

# Size of the episode icon in the download list
IMAGE_SIZE = 64

# Size of the downloads popover
POPOVER_WIDTH = 500
POPOVER_HEIGHT = 400


class DownloadsButton(Gtk.MenuButton):
    """Button displaying the list of downloads

    Signals
    -------
    episode-updated(Episode)
        Emitted at the end the download of an episode.
    """

    __gsignals__ = {
        'episode-updated':
            (GObject.SIGNAL_RUN_FIRST, None, (GObject.TYPE_PYOBJECT,)),
    }

    def __init__(self):
        super().__init__()

        self.logger = logging.getLogger(
            "{}.{}".format(__name__, self.__class__.__name__))

        self.pool = DownloadsPool()

        self.set_image(Gtk.Image.new_from_icon_name(
            "document-save-symbolic", Gtk.IconSize.BUTTON))
        self.set_tooltip_text("Display the progress of ongoing downloads")

        self.popover = Gtk.Popover()

        self.listbox = IndexedListBox()
        self.listbox.set_selection_mode(Gtk.SelectionMode.NONE)
        self.listbox.connect("add", self._on_list_changed)
        self.listbox.connect("remove", self._on_list_changed)
        self.listbox.show()

        scrolled_window = ScrolledWindow()
        scrolled_window.add_with_viewport(self.listbox)
        scrolled_window.show_all()
        scrolled_window.set_vexpand(True)
        scrolled_window.set_size_request(POPOVER_WIDTH, 0)
        try:
            scrolled_window.set_min_content_height(10)
            scrolled_window.set_max_content_height(POPOVER_HEIGHT)
        except AttributeError:
            self.logger.warning(
                "Gtk.ScrolledWindow.set_max_content_height is not available, "
                "using set_size_request.")
            scrolled_window.set_size_request(POPOVER_WIDTH, POPOVER_HEIGHT)
        self.popover.add(scrolled_window)

        self.set_popover(self.popover)
        self.set_sensitive(False)

    def download(self, episode):
        """Add an episode to the download queue"""
        if episode.id in self.listbox.get_ids():
            self.logger.debug("The episode is already in the download list.")
            return

        job = DownloadJob(episode)

        row = DownloadRow(job)
        row.connect("episode-updated", self._on_episode_updated)
        self.listbox.add_with_id(row, episode.id)

        self.pool.add(job)

    def stop(self):
        """Cancel all the downloads"""
        self.pool.stop()

    def _on_list_changed(self, listbox, child):
        """Called when a child is added to or removed from the list"""
        self.set_sensitive(bool(listbox.get_children()))

    def _on_episode_updated(self, row, episode):
        """Called at the end of the download of an episode

        Propagate the signal.
        """
        self.emit("episode-updated", episode)


class DownloadRow(Gtk.ListBoxRow):
    """Row in the download list

    Signals
    -------
    episode-updated
        Emitted at the end of the download of an episode.
    """

    __gsignals__ = {
        'episode-updated':
            (GObject.SIGNAL_RUN_FIRST, None, (GObject.TYPE_PYOBJECT,)),
    }

    def __init__(self, job):
        super().__init__()

        self.job = job
        self.job.connect("start", self._on_start)
        self.job.connect("progress", self._on_progress)
        self.job.connect("remove", self._on_remove)

        # Layout
        self.grid = Gtk.Grid()
        self.grid.set_column_spacing(5)
        self.add(self.grid)

        self.icon = Gtk.Image()
        self.grid.attach(self.icon, 0, 0, 1, 3)

        self.title = Label()
        self.title.set_hexpand(True)
        self.grid.attach(self.title, 1, 0, 1, 1)

        self.progress = Gtk.ProgressBar()
        self.progress.set_valign(Gtk.Align.CENTER)
        self.grid.attach(self.progress, 1, 1, 1, 1)

        self.progress_label = Label()
        self.grid.attach(self.progress_label, 1, 2, 1, 1)

        button = Gtk.Button.new_from_icon_name("edit-delete-symbolic",
                                               Gtk.IconSize.BUTTON)
        button.set_relief(Gtk.ReliefStyle.NONE)
        button.connect('clicked', self._on_cancel_clicked)
        self.grid.attach(button, 2, 0, 1, 3)

        self.show_all()

        self.set_pending()

    def _on_start(self, job):
        """Update the widget to show that the download is starting"""
        self.progress_label.set_text("Starting...")

    def _on_progress(self, job, currentsize, totalsize, speed):
        """Update the widget to show the current download progress"""
        try:
            remaining_time = int(round((totalsize - currentsize) / speed))
        except ZeroDivisionError:
            remaining_time = None

        self.progress.set_fraction(currentsize / totalsize)
        if remaining_time is None:
            self.progress_label.set_text("{} of {} ({}/s)".format(
                format_size(currentsize), format_size(totalsize),
                format_size(speed)
            ))
        else:
            self.progress_label.set_text(
                "{} - {} of {} ({}/s)".format(
                    format_fulltext_duration(remaining_time),
                    format_size(currentsize), format_size(totalsize),
                    format_size(speed)
                ))

    def _on_remove(self, job):
        """Remove the download from the list"""
        self.emit("episode-updated", self.job.episode)
        self.destroy()

    def set_pending(self):
        """Update the widget to show the download is pending"""
        self.icon.set_from_pixbuf(
            self.job.episode.podcast.image.as_pixbuf(IMAGE_SIZE))

        self.title.set_markup(
            "{} - <i>{}</i>".format(
                GLib.markup_escape_text(self.job.episode.title),
                GLib.markup_escape_text(self.job.episode.podcast.title)))

        self.progress_label.set_text("Pending...")

    def _on_cancel_clicked(self, button):
        """Called when the cancel button is clicked"""
        self.job.cancel()


class DownloadJob(GObject.Object):
    """Download job that should be added to a DownloadsPool

    Signals
    -------
    start
        Emitted when the download starts.
    remove
        Emitted when the download should be removed from the list (e.g. removed
        or canceled).
    progress(current_size, total_size, speed)
        Emitted during the download to report its progress.
    """

    __gsignals__ = {
        'start':
            (GObject.SIGNAL_RUN_FIRST, None, ()),
        'remove':
            (GObject.SIGNAL_RUN_FIRST, None, ()),
        'progress':
            (GObject.SIGNAL_RUN_FIRST, None, (int, int, float)),
    }

    def __init__(self, episode):
        GObject.Object.__init__(self)

        self.episode = episode
        self.canceled = False

    def cancel(self):
        """Cancel the download"""
        self.canceled = True
        GObject.idle_add(self.emit, "remove")

    def start(self):
        """Start the download"""
        GObject.idle_add(self.emit, "start")

        generator = download_with_average_speed(self.episode)

        for current_size, file_size, average_speed in generator:
            if self.canceled:
                generator.close()
                return

            GObject.idle_add(self.emit, "progress", current_size, file_size,
                             average_speed)

        GObject.idle_add(self.emit, "remove")



# -*- coding: utf-8 -*-
#
# Copyright (c) 2016 Muges
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation files
# (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge,
# publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""
An horizontal Gtk.Paned which becomes vertical if there is not enough width
"""

from gi.repository import Gtk


class Paned(Gtk.Paned):
    """An horizontal Gtk.Paned which becomes vertical if there is not enough
    width"""
    def __init__(self):
        super().__init__()

        self.limit_width = 0

        self.connect('size_allocate', Paned._on_size_allocate)

    def set_limit_width(self, width):
        """Set the minimal width of the widget

        If the available width is less than the minimal width, the
        Paned will be reoriented vertically."""
        self.limit_width = width

    def _on_size_allocate(self, allocation):
        if allocation.width < self.limit_width:
            self.set_orientation(Gtk.Orientation.VERTICAL)
        else:
            self.set_orientation(Gtk.Orientation.HORIZONTAL)



# -*- coding: utf-8 -*-
#
# Copyright (c) 2016 Muges
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation files
# (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge,
# publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import os
import logging
import threading
from gi.repository import Gio
from gi.repository import GLib
from gi.repository import GObject
from gi.repository import Gtk

from erika import library
from erika.config import DEFAULT_CONFIG_DIRECTORY
from erika.library.models import Config, Episode, Podcast
from erika.library.opml import import_opml, export_opml
from erika.library.gpodder import GPodderClient, GPodderUnauthorized
from erika.util import check_connection
from . import preferences
from .main_window import MainWindow
from .util import cb, get_builder


class Application(Gtk.Application):
    """
    Signals
    -------
    network-state-changed(online, error)
        Emitted when the network state changes
    """
    __gsignals__ = {
        'network-state-changed':
            (GObject.SIGNAL_RUN_FIRST, None, (GObject.TYPE_BOOLEAN,
                                              GObject.TYPE_BOOLEAN)),
    }

    def __init__(self):
        Gtk.Application.__init__(
            self, application_id="fr.muges.erika",
            flags=Gio.ApplicationFlags.HANDLES_COMMAND_LINE)

        self.logger = logging.getLogger(
            "{}.{}".format(__name__, self.__class__.__name__))

        self.window = None
        self.online = True
        self.configuration_directory = DEFAULT_CONFIG_DIRECTORY
        self.synchronization_lock = threading.Lock()

        self.add_main_option("offline", ord("o"), GLib.OptionFlags.NONE,
                             GLib.OptionArg.NONE, "Enable offline mode", None)

        self.add_main_option(
            "configuration-directory", ord("c"), GLib.OptionFlags.NONE,
             GLib.OptionArg.STRING, "Path of the configuration directory.",
             None)

        self.connect("startup", cb(self._on_startup))
        self.connect("activate", cb(self._on_activate))
        self.connect("command-line", cb(self._on_command_line))
        self.connect("shutdown", cb(self.close_main_window))
        self.connect("shutdown", cb(self._on_shutdown))

    def _on_startup(self):
        """Called when the first instance of the application is launched"""
        # pylint: disable=attribute-defined-outside-init

        Gtk.Application.do_startup(self)

        # Creates the actions
        self.add_podcast_action = Gio.SimpleAction.new("add-podcast", None)
        self.add_podcast_action.connect("activate", cb(self.add_podcast, 2))
        self.add_action(self.add_podcast_action)

        action = Gio.SimpleAction.new("import-opml", None)
        action.connect("activate", cb(self.import_opml, 2))
        self.add_action(action)

        action = Gio.SimpleAction.new("export-opml", None)
        action.connect("activate", cb(self.export_opml, 2))
        self.add_action(action)

        self.update_action = Gio.SimpleAction.new("update", None)
        self.update_action.connect("activate", cb(self.synchronize_library, 2))
        self.add_action(self.update_action)

        self.sync_action = Gio.SimpleAction.new("force-synchronization", None)
        self.sync_action.connect("activate", cb(self.force_synchronization, 2))
        self.add_action(self.sync_action)

        action = Gio.SimpleAction.new("preferences", None)
        action.connect("activate", cb(preferences.run, 2), self.window)
        self.add_action(action)

        action = Gio.SimpleAction.new("quit", None)
        action.connect("activate", cb(self.quit, 2))
        self.add_action(action)
        self.add_accelerator("<Primary>q", "app.quit", None)

        if self.prefers_app_menu():
            # Create the menu
            builder = get_builder('data/menu.glade')
            self.set_app_menu(builder.get_object("app-menu"))

        # Set windows icons
        Gtk.Window.set_default_icon_name('erika')

    def _on_activate(self):
        """Called when the application is launched"""
        if self.window is not None:
            self.logger.info('Erika is already running.')
            return

        # Setup logging
        logger = logging.getLogger("erika")
        logger.setLevel(logging.DEBUG)
        logger.propagate = False

        formatter = logging.Formatter(
            '%(levelname)-8s (%(name)s) : %(message)s')

        # Display logs on stdout
        handler = logging.StreamHandler()
        handler.setLevel(logging.DEBUG)
        handler.setFormatter(formatter)
        logger.addHandler(handler)

        # Create the configuration directory if it does not exists
        if not os.path.isdir(self.configuration_directory):
            os.makedirs(self.configuration_directory)

        # Save logs
        logfile = os.path.join(self.configuration_directory, 'debug.log')
        handler = logging.FileHandler(logfile, mode='w')
        handler.setLevel(logging.DEBUG)
        handler.setFormatter(formatter)
        logger.addHandler(handler)

        # Connect to the database
        library.initialize(self.configuration_directory)

        # Create the main window
        try:
            self.window = MainWindow(self)
        except BaseException:
            self.logger.exception("Unable to create main window.")

        self.set_network_state(self.online)

        # Update the library
        interval = Config.get_value("library.synchronize_interval")

        self.synchronize_library(scan=True)
        GObject.timeout_add_seconds(interval * 60, self.synchronize_library)

        self.window.show_all()
        self.window.present()

    def _on_command_line(self, command_line):
        """Called when the application is launched from the command line"""
        if self.window is not None:
            self.logger.info('Erika is already running.')
            return 0

        options = command_line.get_options_dict()

        if options.contains("offline"):
            self.logger.info("Started in offline mode.")
            self.online = False

        configuration_directory = options.lookup_value('configuration-directory', None)
        if (configuration_directory and
            configuration_directory.is_of_type(GLib.VariantType.new("s"))):
            self.configuration_directory = configuration_directory.get_string()

        self.activate()
        return 0

    def _on_shutdown(self):
        """Called when the application is closed"""
        if self.window is None:
            return

        self.close_main_window()

        Episode.update(new=False).execute()
        self.synchronize_library(update=False)

        Gtk.Application.do_shutdown(self)

    def _synchronization_end(self, message_id):
        """Called at the end of the synchronization"""
        self.window.statusbox.remove(message_id)

        self.window.podcast_list.update()
        self.window.episode_list.update()
        self.window.update_counts()

    def _synchronization_thread(self, message_id, update=True, scan=False):
        acquired = self.synchronization_lock.acquire(blocking=False)
        if not acquired:
            # There is already another synchronization thread running : abort
            return

        try:
            client = GPodderClient()

            if not self.online:
                return

            # Check internet connection
            hostname = Config.get_value("network.connection_check_hostname")
            if not check_connection(hostname):
                self.set_network_state(False, error=True)
                return

            GObject.idle_add(self.window.statusbox.add,
                             "Synchronizing subscriptions...", message_id)

            try:
                client.connect()
            except GPodderUnauthorized:
                self.window.error_label.set_text(
                    'Unable to connect to gpodder: invalid credentials.')
                self.window.error_bar.show()
            except Exception:
                self.window.error_label.set_text(
                    'Unable to connect to gpodder.')
                self.window.error_bar.show()
            else:
                self.window.error_bar.hide()

            client.synchronize_subscriptions()

            if update:
                GObject.idle_add(self.window.statusbox.edit,
                                 message_id, "Updating library...")

                for podcast in Podcast.select():
                    podcast.update_podcast()
                    GObject.idle_add(
                        self.window.podcast_list.update_podcast, podcast)

            if update and scan:
                GObject.idle_add(self.window.statusbox.edit,
                                 message_id, "Scanning library...")
                library.scan()

            GObject.idle_add(self.window.statusbox.edit,
                             message_id, "Synchronizing episode actions...")
            if update:
                client.synchronize_episode_actions()
            else:
                client.push_episode_actions()
        finally:
            GObject.idle_add(self._synchronization_end, message_id)
            self.synchronization_lock.release()

    def synchronize_library(self, update=True, scan=False):
        """Synchronize the podcasts library with gpodder.net"""
        message_id = self.window.statusbox.get_next_message_id()

        thread = threading.Thread(target=self._synchronization_thread,
                                  args=(message_id, update, scan))
        thread.start()

        return True

    def force_synchronization(self):
        """Synchronize the podcasts library with gpodder.net, taking into
        account every episode action and every subscriptions changes"""
        Config.set_value("gpodder.last_subscription_sync", "0")
        Config.set_value("gpodder.last_episodes_sync", "0")

        self.synchronize_library()

    def add_podcast(self):
        """
        Add a new podcast.
        """
        # Open the "Add a podcast" dialog
        builder = get_builder('data/add_podcast.glade')
        dialog = builder.get_object("add_dialog")
        url_entry = builder.get_object("url_entry")

        dialog.set_transient_for(self.window)

        response = dialog.run()
        url = url_entry.get_text()
        dialog.destroy()

        if response != Gtk.ResponseType.OK or not url:
            return

        message_id = self.window.statusbox.get_next_message_id()
        self.window.statusbox.add("Adding podcast...", message_id)

        def _end():
            self.window.statusbox.remove(message_id)

            self.window.podcast_list.update()
            self.window.episode_list.update()
            self.window.update_counts()

        def _add(url):
            # TODO : handle errors
            podcast = Podcast.new("rss", url)
            podcast.update_podcast()

            GObject.idle_add(_end)

        thread = threading.Thread(target=_add, args=(url,))
        thread.start()

    def import_opml(self):
        """
        Import the podcasts subscriptions from an OPML file.
        """
        dialog = Gtk.FileChooserDialog(
            "Import OPML file", self.window,
            Gtk.FileChooserAction.OPEN,
            (Gtk.STOCK_CANCEL, Gtk.ResponseType.CANCEL,
             Gtk.STOCK_OPEN, Gtk.ResponseType.OK))

        # Add filters
        filter_xml = Gtk.FileFilter()
        filter_xml.set_name("XML files")
        filter_xml.add_mime_type("text/xml")
        dialog.add_filter(filter_xml)

        filter_text = Gtk.FileFilter()
        filter_text.set_name("Text files")
        filter_text.add_mime_type("text/plain")
        dialog.add_filter(filter_text)

        filter_any = Gtk.FileFilter()
        filter_any.set_name("Any files")
        filter_any.add_pattern("*")
        dialog.add_filter(filter_any)

        response = dialog.run()
        if response == Gtk.ResponseType.OK:
            filename = dialog.get_filename()
        elif response == Gtk.ResponseType.CANCEL:
            filename = None
        dialog.destroy()

        if not filename:
            return

        message_id = self.window.statusbox.get_next_message_id()
        self.window.statusbox.add("Importing OPML...", message_id)

        def _end():
            self.window.statusbox.remove(message_id)

            self.window.podcast_list.update()
            self.window.episode_list.update()
            self.window.update_counts()

        def _import(filename):
            # TODO : handle errors
            import_opml(filename)

            GObject.idle_add(_end)

        thread = threading.Thread(target=_import, args=(filename,))
        thread.start()

    def export_opml(self):
        """
        Export the podcasts subscriptions in an OPML file.
        """
        dialog = Gtk.FileChooserDialog(
            "Export OPML file", self.window,
            Gtk.FileChooserAction.SAVE,
            (Gtk.STOCK_CANCEL, Gtk.ResponseType.CANCEL,
             Gtk.STOCK_SAVE, Gtk.ResponseType.OK))

        response = dialog.run()
        if response == Gtk.ResponseType.OK:
            filename = dialog.get_filename()
        elif response == Gtk.ResponseType.CANCEL:
            filename = None
        dialog.destroy()

        if not filename:
            return

        message_id = self.window.statusbox.get_next_message_id()
        self.window.statusbox.add("Exporting OPML...", message_id)

        def _end():
            self.window.statusbox.remove(message_id)

        def _export(filename):
            export_opml(filename)

            GObject.idle_add(_end)

        thread = threading.Thread(target=_export, args=(filename,))
        thread.start()

    def set_network_state(self, online=True, error=False):
        if online and not error:
            hostname = Config.get_value("network.connection_check_hostname")
            error = not check_connection(hostname)
        if error:
            online = False

        self.online = online

        self.add_podcast_action.set_enabled(online)
        self.update_action.set_enabled(online)
        self.sync_action.set_enabled(online)

        self.emit('network-state-changed', online, error)

    def get_online(self):
        """Return True if the application is connected to the internet"""
        return self.online

    def close_main_window(self):
        if self.window is not None:
            self.window.destroy()



# -*- coding: utf-8 -*-
#
# Copyright (c) 2016 Muges
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation files
# (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge,
# publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""
The :mod:`erika.downloads` module provides a dowloading function with average
speed computation, as well as a :class:`DownloadPool` to implement parallel
downloading.
"""

import logging
import os
from queue import Queue, Empty
from threading import Thread
import time
import requests

from . import tags
from .library.models import Config, EpisodeAction
from .util import guess_extension

# Smoothing factor used to compute the average download speed
SMOOTHING_FACTOR = 0.01

# The time step used to compute the average download speed, in seconds
TIME_DELTA = 0.05


def download_chunks(episode):
    """Download an episode by chunks.

    Yields
    ------
    int, int
        The size of the previous chunk, and the total size of the file.
    """
    logger = logging.getLogger(__name__)

    # Start the download and get the mimetype of the file
    logger.debug("Getting file mimetype.")
    response = requests.get(episode.file_url, stream=True)
    mimetype = (
        response.headers.get('content-type') or episode.mimetype
    )
    try:
        file_size = int(response.headers.get('content-length'))
    except TypeError:
        file_size = None

    # Set the extension of the file
    ext = guess_extension(mimetype)

    filename = episode.get_default_filename(ext)
    tempfilename = filename + ".part"
    dirname = os.path.dirname(filename)

    # Create the podcast directory if necessary
    if not os.path.isdir(dirname):
        os.makedirs(dirname)

    logger.debug("Downloading to temporary file '%s'.", tempfilename)
    try:
        # Download to a temporary file
        with open(tempfilename, "wb") as fileobj:
            if file_size is None:
                fileobj.write(response.content)
            else:
                for data in response.iter_content(chunk_size=1024):
                    # Write the downloaded data to the temporary file
                    fileobj.write(data)
                    chunk_size = len(data)

                    yield (chunk_size, file_size)

        # Move the temporary file to the destination file
        os.rename(tempfilename, filename)

        # Set the tags of the downloaded file
        tags.set_tags(filename, episode)

        episode.absolute_local_path = filename
        episode.save()

        action = EpisodeAction(episode=episode, action="download")
        action.save()
    finally:
        # Remove the temporary file if it exists (if the download failed or was
        # canceled)
        if os.path.isfile(tempfilename):
            os.remove(tempfilename)


def download_with_average_speed(episode):
    """Download an episode by chunks, and compute the average speed (see
    https://en.wikipedia.org/wiki/Moving_average#Exponential_moving_average).

    Yields
    ------
    int, int, float
        - The current downloaded size in bytes;
        - the total file size in bytes;
        - the average downloading speed in bytes per second.
    """
    previous_time = time.time()  # Time of the previous step
    current_size = 0  # Total downloaded size
    step_size = 0  # Size downloaded during the current time step
    average_speed = 0  # Average download speed

    for chunk_size, file_size in download_chunks(episode):
        now = time.time()

        current_size += chunk_size
        step_size += chunk_size

        if now - previous_time >= TIME_DELTA:
            # Compute the average speed during current time step
            current_speed = step_size / (now - previous_time)

            # Compute the smoothed average download speed
            if average_speed == 0:
                average_speed = current_speed
            else:
                average_speed = (
                    SMOOTHING_FACTOR * current_speed +
                    (1 - SMOOTHING_FACTOR) * average_speed)

            yield (current_size, file_size, average_speed)

            step_size = 0
            previous_time = now


class DownloadsPool(object):
    """Download pool, handling the downloads queue and the workers to make
    concurrent downloads possible."""
    def __init__(self):
        self.queue = Queue()

        workers = Config.get_value("downloads.workers")

        self.workers = [
            DownloadWorker(self.queue) for _ in range(0, workers)
        ]

    def add(self, job):
        """Add a new job to the queue.

        A job should have :

        - a ``cancel`` method, which stops the job cleanly and as soon as
          possible;
        - a ``start`` method, which starts the job.

        See :class:`erika.frontend.downloads.DownloadJob` for an example.
        """
        self.queue.put(job)

    def stop(self):
        """Stop all the workers (and the jobs being currently executed)."""
        for worker in self.workers:
            worker.stop()


class DownloadWorker(Thread):
    """Thread waiting for jobs to be added to the queue and executing them."""
    def __init__(self, queue):
        Thread.__init__(self)

        self.logger = logging.getLogger(
            "{}.{}".format(__name__, self.__class__.__name__))

        self.queue = queue
        self.stopped = False
        self.current_job = None

        self.start()

    def run(self):
        while not self.stopped:
            try:
                self.current_job = self.queue.get(timeout=1)
            except Empty:
                continue

            try:
                self.current_job.start()
            except Exception:  # pylint: disable=broad-except
                self.logger.exception("Download failed.")
            finally:
                self.current_job = None
                self.queue.task_done()

    def stop(self):
        """Stop the worker (and the job being currently executed)."""
        self.stopped = True
        if self.current_job:
            self.current_job.cancel()
            self.current_job = None
